local ADDON_NAME="AVR"
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)

local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local max = math.max
local min = math.min
local abs = math.abs
local atan2 = math.atan2
local sin = math.sin
local cos = math.cos
local type = type
local select = select
local next = next
local GetTime = GetTime
local deg2rad = math.pi/180.0
local pi = math.pi
local tostring,tonumber=tostring,tonumber
local strlen=string.len

local Core=AVR

AVRBigWigsModule=Core:NewModule("AVRBigWigsModule","AceEvent-3.0","AceTimer-3.0")

function AVRBigWigsModule:OnInitialize()
end

function AVRBigWigsModule:OnEnable()
	self:RegisterMessage("BigWigs_Message")
end

function AVRBigWigsModule:OnDisable()
	self:UnregisterMessage("BigWigs_Message")	
end

local sceneObject=nil
function AVRBigWigsModule:BigWigs_Message(...)
	if sceneObject then sceneObject:BigWigs_Message(...) end
end

local stripColor=function(s)
	local ind=s:find("|c",1,true)
	if ind==nil then return s end
	local ind2=s:find("|r",ind+2,true)
	if ind2==nil then return s end
	return s:sub(ind+10,ind2-1)
end

local onYouPattern = nil
local onOtherPattern = nil

local getTargetFromMessage=function(text)
	if not onYouPattern then
		local BWL = LibStub("AceLocale-3.0"):GetLocale("Big Wigs: Common")
		if not BWL then return nil end
		onYouPattern = BWL["you"]:gsub("%%s","(.+)")
		onOtherPattern = BWL["other"]:gsub("%%s","(.+)")
	end
	
	local target=nil
	if text:match(onYouPattern) then
		return "player"
	else
		local _,target=text:match(onOtherPattern)
		if target then return stripColor(target)
		else return nil end
	end
end

AVRBigWigsScene={Embed=Core.Embed}

AVRBigWigsScene.class="AVRBigWigsScene"
function AVRBigWigsScene:New(threed)
	if self ~= AVRBigWigsScene then return end
	
	if sceneObject ~=nil then
		Core:Print(L["You can add only one BigWigs scene"])
		return nil
	end
	
	local s=AVRScene:New(threed)
	AVRBigWigsScene:Embed(s)
	s.class=AVRBigWigsScene.class
	s.autoFireOptionChanges=false
	s.name=L["BigWigs scene"]
	s.options={
		IcecrownCitadel={
			ProfessorPutricide={
				malleableGoo=true
			},
			TheLichKing={
				defile=true
			}
		}
	}
	sceneObject=s
	
	return s
end

function AVRBigWigsScene:OnRemove()
	sceneObject=nil
end

function AVRBigWigsScene:BigWigs_Message(message, module, key, text, color, _, sound, broadcastOnly, icon)
	if key==72295 then -- malleable goo
		if broadcastOnly then return end
		local target=getTargetFromMessage(text)
		
		if target then
			local px,py=self.threed:GetUnitPosition(target)
			if px~=0 or py~=0 then
				local m=AVRFilledCircleMesh:New(5)
				m:SetColor(1.0,0.0,0.0,0.5)
				m:TranslateMesh(px,py)
				self:AddMesh(m)
				AVRBigWigsModule:ScheduleTimer(function()
					m:Remove()
				end,5)
			end
		end
	elseif key==72762 then -- defile
		if broadcastOnly then return end
		local target=getTargetFromMessage(text)
		
		if target then
			local px,py=self.threed:GetUnitPosition(target)
			if px~=0 or py~=0 then
				local m=AVRFilledCircleMesh:New(5)
				m:SetColor(1.0,0.0,0.0,0.5)
				m.followUnit=target
				self:AddMesh(m)
				AVRBigWigsModule:ScheduleTimer(function()
					m:Remove()
				end,2)
			end
		end
	elseif key==70126 then -- frost beacon
		local target=getTargetFromMessage(text)
		local targets={}
		local pos=strfind(target,",")
		while pos do
			local t=target:sub(1,pos-1):gsub("^%s*",""):gsub("%s$","")
			insert(targets,t)
			target=target:sub(pos+1)
			pos=strfind(target,",")
		end
		local meshes={}
		for i,t in ipairs(targets) do
			local m=AVRFilledCircleMesh:New(5)
			m:SetColor(1.0,0.0,0.0,0.5)
			m.followUnit=t
			self:AddMesh(m)
			meshes[#meshes+1]=m
		end
		AVRBigWigsModule:ScheduleTimer(function()
			for i,m in ipairs(meshes) do
				m:Remove()
			end
		end,7)
	elseif key==69279 then -- festergut spores
		local target=getTargetFromMessage(text)
		local targets={}
		local pos=strfind(target,",")
		while pos do
			local t=target:sub(1,pos-1):gsub("^%s*",""):gsub("%s$","")
			insert(targets,t)
			target=target:sub(pos+1)
			pos=strfind(target,",")
		end
		local meshes={}
		for i,t in ipairs(targets) do
			local m=AVRFilledCircleMesh:New(8)
			m:SetColor(1.0,0.0,0.0,0.5)
			m.followUnit=t
			self:AddMesh(m)
			meshes[#meshes+1]=m
		end
		AVRBigWigsModule:ScheduleTimer(function()
			for i,m in ipairs(meshes) do
				m:Remove()
			end
		end,7)
	end
end

function AVRBigWigsScene:GetOptions()
	return {
		type = "group",
		name = self.name,	
		get = function(info) 
			if type(info)~="string" then info=info[#info] end
			local func="Get"..string.upper(string.sub(info,1,1))..string.sub(info,2)
			if self[func] then return self[func](self) 
			else return self[info] end
		end,
		set = function(info,val)
			if type(info)~="string" then info=info[#info] end
			local func="Set"..string.upper(string.sub(info,1,1))..string.sub(info,2)
			if self[func] then self[func](self,val)
			else self[info]=val end
		end,
		args = {
			name = {
				type = "input",
				name = L["Name"],
				order = 10,
				width = "full",
				set = 	function(_,val)
							self.name=val
							Core:OptionsChanged()
						end
			},
			visible = {
				type = "toggle",
				name = L["Visible"],
				order = 15
			},
			select = {
				type = "toggle",
				name = L["Select"],
				order = 20,
				get =	function(_)
							if Core.sceneManager:GetSelectedScene()==self then return true
							else return false end
						end,
				set =	function(_,val)
							if val then Core.sceneManager:SetSelectedScene(self)
							else Core.sceneManager:SetSelectedScene(nil) end
						end
			},
			remove = {
				type = "execute",
				name = L["Remove scene"],
				order = 30,
				width = "full",
				confirm = true,
				func = function() self:Remove() end
			},
		}
	}
end

function AVRBigWigsScene:Pack()
	local temp=self.meshes
	self.meshes={}
	local s=AVRScene.Pack(self)


	self.meshes=temp
	return s
end

function AVRBigWigsScene:Unpack(s)
	AVRScene.Unpack(self,s)



end

AVR:RegisterSceneClass(AVRBigWigsScene,L["BigWigs scene"],false)
