local ADDON_NAME="AVR"
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)

local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local max = math.max
local min = math.min
local abs = math.abs
local atan2 = math.atan2
local sin = math.sin
local cos = math.cos
local type = type
local select = select
local next = next
local GetTime = GetTime
local deg2rad = math.pi/180.0
local pi = math.pi
local tostring,tonumber=tostring,tonumber
local strlen=string.len

local Core=AVR

local raidID={}
local partyID={}
for i=1,40 do raidID[i]="raid"..i end
for i=1,4 do partyID[i]="party"..i end


AVRRangeWarningScene={Embed=Core.Embed}
AVRRangeWarningScene.class="AVRRangeWarningScene"
function AVRRangeWarningScene:New(threed,range,radius)
	if self ~= AVRRangeWarningScene then return end
	local s=AVRScene:New(threed)
	AVRRangeWarningScene:Embed(s)
	s.class=AVRRangeWarningScene.class
	s.range=range or 10
	s.radius=radius or 1
	s.rangeSq=s.range*s.range
	s.spokes=6
	s.autoFireOptionChanges=false
	s.name=L["Range warning"]
	return s
end

function AVRRangeWarningScene:SetRange(range)
	self.range=range
	self.rangeSq=self.range*self.range
	self:ClearScene()
end

function AVRRangeWarningScene:SetRadius(radius)
	self.radius=radius
	self:ClearScene()
end

function AVRRangeWarningScene:GetOptions()
	return {
		type = "group",
		name = self.name,	
		get = function(info) 
			if type(info)~="string" then info=info[#info] end
			local func="Get"..string.upper(string.sub(info,1,1))..string.sub(info,2)
			if self[func] then return self[func](self) 
			else return self[info] end
		end,
		set = function(info,val)
			if type(info)~="string" then info=info[#info] end
			local func="Set"..string.upper(string.sub(info,1,1))..string.sub(info,2)
			if self[func] then self[func](self,val)
			else self[info]=val end
		end,
		args = {
			name = {
				type = "input",
				name = L["Name"],
				order = 10,
				width = "full",
				set = 	function(_,val)
							self.name=val
							Core:OptionsChanged()
						end
			},
			visible = {
				type = "toggle",
				name = L["Visible"],
				order = 15
			},
			select = {
				type = "toggle",
				name = L["Select"],
				order = 20,
				get =	function(_)
							if Core.sceneManager:GetSelectedScene()==self then return true
							else return false end
						end,
				set =	function(_,val)
							if val then Core.sceneManager:SetSelectedScene(self)
							else Core.sceneManager:SetSelectedScene(nil) end
						end
			},
			remove = {
				type = "execute",
				name = L["Remove scene"],
				order = 30,
				width = "full",
				confirm = true,
				func = function() self:Remove() end
			},
			range = {
				type = "range",
				name = L["Range"],
				order = 40,
				width = "full",
				min = 0.0, max=500, bigStep=1.0
			},
			radius = {
				type = "range",
				name = L["Radius"],
				order = 50,
				width = "full",
				min = 0.0, max=10, bigStep=0.5
			}
		}
	}
end

local markers={}
for i=1,40 do insert(markers,{}) end
function AVRRangeWarningScene:DrawScene()
	local px,py
	local sx,sy=self.threed:GetUnitPosition("player")
	local dx,dy
	local numMarkers=0
	if GetNumRaidMembers()>0 then
		for i=1,40 do
			if not UnitIsUnit(raidID[i],"player") then 
				px,py=self.threed:GetUnitPosition(raidID[i])
				if px~=0 or py~=0 then
					dx,dy=sx-px,sy-py
					if dx*dx+dy*dy<=self.rangeSq then
						numMarkers=numMarkers+1
						markers[numMarkers].x,markers[numMarkers].y=px,py
					end
				end
			end
		end
	else
		for i=1,4 do
			px,py=self.threed:GetUnitPosition(partyID[i])
			if px~=0 or py~=0 then
				dx,dy=sx-px,sy-py
				if dx*dx+dy*dy<=self.rangeSq then
					numMarkers=numMarkers+1
					markers[numMarkers].x,markers[numMarkers].y=px,py
				end
			end
		end
	end
	while #self.meshes<numMarkers do 
		local m=self:AddMesh(AVRMarkerMesh:New(self.radius,self.spokes))
		m.followPlayer=false
		m.followRotation=false
	end
	for i=1,numMarkers do
		self.meshes[i]:OnUpdate(self.threed)
		self.meshes[i].translateX=markers[i].x
		self.meshes[i].translateY=markers[i].y
		self.threed:DrawMesh(self.meshes[i])
	end
end

function AVRRangeWarningScene:Pack()
	self:ClearScene()
	local s=AVRScene.Pack(self)
	s.radius=self.radius
	s.range=self.range
	s.spokes=self.spokes
	return s
end

function AVRRangeWarningScene:Unpack(s)
	AVRScene.Unpack(self,s)
	self.range=s.range or 10
	self.rangeSq=self.range*self.range
	self.radius=s.radius or 1
	self.spokes=s.spokes or 6
	self:ClearScene()
end

AVR:RegisterSceneClass(AVRRangeWarningScene,L["Range Warning"],true)
