local ADDON_NAME="AVR"
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)

local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local max = math.max
local min = math.min
local abs = math.abs
local atan2 = math.atan2
local sin = math.sin
local cos = math.cos
local type = type
local select = select
local next = next
local GetTime = GetTime
local deg2rad = math.pi/180.0
local pi = math.pi

local Core=AVR

AVRSceneManager = {Embed = Core.Embed}
local T=AVRSceneManager

function T:New(threed)
	if self ~= T then return end
	local s={}
	
	T:Embed(s)

	s.threed=threed
	s.scenes={}
	s.sceneid=0
	s.selection=0
	
	s.scenesLoaded=false
	
	return s
end

function T:LoadScenes()
	self.scenes={}
	if not Core.db.profile.scenes then 
		self.sceneid=1
		self.scenesLoaded=true
		return
	end
	self.sceneid=Core.db.profile.sceneid or 1
	for i,scene in ipairs(Core.db.profile.scenes) do
		local s=self:UnpackScene(scene,false)
		self.scenes[i]=s
		s:SetSceneManager(self)
	end
	for i,scene in ipairs(self.scenes) do
		if scene.from==nil and self.sceneid<=scene.id then
			self.sceneid=scene.id+1
		end
	end
	
	local ids={}
	for i,scene in ipairs(self.scenes) do
		if scene.from==nil then
			if not scene.id or scene.id==0 or ids[scene.id] then
				Core:Print(L["Loaded scene had invalid scene id, reassigning"])
				scene.id=self:GetNewSceneID()
			end
		end
	end
	self.selection=Core.db.profile.selectedScene
	if not self.selection then self.selection=0 end
	
	self.scenesLoaded=true
	Core:OptionsChanged()
end

function T:GetSelectedScene()
	if self.selection<=0 or self.selection>#self.scenes then return nil
	else return self.scenes[self.selection] end
end

function T:SetSelectedScene(scene)
	if scene~=nil then
		if self.selection>0 and self.scenes[self.selection]==scene then return true end
		for i,s in ipairs(self.scenes) do
			if s==scene then
				self.selection=i
				Core:OptionsChanged()
				return true
			end
		end
	end
	self.selection=0
	Core:OptionsChanged()
	return false
end

function T:SaveScenes()
	-- don't save if we didn't load anything, otherwise we'll clear all scenes
	-- whenever there's a bug somewhere in loading
	if not self.scenesLoaded then return end
	
	Core.db.profile.scenes={}
	Core.db.profile.sceneid=self.sceneid
	Core.db.profile.selectedScene=self.selection
	for i,scene in ipairs(self.scenes) do
		Core.db.profile.scenes[i]=scene:Pack()
	end
end

function T:DrawScenes()
	local scene
	local currentZone=Core.zoneData:GetCurrentZone()
	for i=1,#self.scenes do
		scene=self.scenes[i]
		if scene.visible and (scene.zone==nil or scene.zone==currentZone) then 
			self.scenes[i]:DrawScene()
		end
	end
end

function T:GetNewSceneID()
	local id=self.sceneid
	self.sceneid=self.sceneid+1
	return id
end

function T:UnpackScene(scene,commReceive)
	local class=scene.class
	if not class then class="AVRScene" end
	if AVR.sceneClasses[class] then
		if not commReceive or AVR.receiveSceneClasses[class] then
			local s=AVR.sceneClasses[class]:New(self.threed)
			s:Unpack(scene)
			return s
		else
			Core:Print(format(L["Received scene with class %s but that class is not receivable"],class))
		end
	else
		Core:Print(format(L["Trying to unpack scene, unknown class %s"],class))
	end
	return nil
end

function T:ReceiveScene(scene)
	scene:SetSceneManager(self)
	for i=1,#self.scenes do
		local s=self.scenes[i]
		if s.from==scene.from and s.id==scene.id then
			self.scenes[i]=scene
			return
		end
	end
	insert(self.scenes,scene)
	
	Core:OptionsChanged()
end

function T:AddScene(scene)
	scene.from=nil
	scene.id=self:GetNewSceneID()
	scene:SetSceneManager(self)
	insert(self.scenes,scene)
	self.selection=#self.scenes
	
	Core:OptionsChanged()	
end

function T:RemoveScene(scene)
	for i=1,#self.scenes do
		local s=self.scenes[i]
		if s==scene then
			remove(self.scenes,i)
			scene:OnRemove()
			break
		end
	end
	
	Core:OptionsChanged()
end
